/* This is for linux 2.2.x (tested with 2.2.1[67])
 * Copyright (C) 2000 Salvatore Sanfilippo <antirez@linuxcare.com>
 *
 * WARNING: You need to patch the kernel using the patch sigsegv.patch
 * compile with:
 * gcc -O2 -Wall -D__KERNEL__ -DMODULE sigsegv.c -c
 *
 * usage: echo [01] > /proc/sys/sigsegv/status
 */

/* __SMP__ auto definition */
#include <linux/autoconf.h>
#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>

#include <asm/uaccess.h>
#include <asm/page.h>

#include "ssversion.h"
#include "sscommon.h"

/* Don't change this, remember whe have just one page */
#define HISTORY_SIZE PAGE_SIZE/ENTRY_OUTPUT_SIZE

/* This is the hook, called in kernel/signal.c */
extern int (*sigsegv_hook) (int sig, struct siginfo *info, struct task_struct *t);

/* array to store the last SEGFAULTED_SIZE processes that
 * exited on signal 11 */
static struct struct_sigsegv {
	char comm[32];
	unsigned long address;
	unsigned int error_code;
	pid_t pid;
	time_t timestamp;
	uid_t uid;
	uid_t pad;
} history[HISTORY_SIZE];

/* the index of the circular buffer */
static int sf_index = 0;
static struct proc_dir_entry *sigsegv_root = NULL;
static struct proc_dir_entry *status_entry = NULL;
static struct proc_dir_entry *history_entry = NULL;
static spinlock_t lock = SPIN_LOCK_UNLOCKED;

/* WARNING: If you don't want to stop the send_sig_info() function this function
 * MUST ever return SIGHOOK_CONTINUE, otherwise send_sig_info will set the
 * error code to the function exit code, unlock the spinlock and return. */
static int log_sigsegv(int sig, struct siginfo *info, struct task_struct *t)
{
	/* we trap only the SIGSEGV */
	if (sig != SIGSEGV)
		return SIGHOOK_CONTINUE;

	printk(KERN_NOTICE \
		"pid %d (%s), uid %d exited on signal 11 %s 0x%08lx\n",
		t->pid, t->comm, t->uid,
		(t->tss.error_code & 2) ?
			"writing" : "reading or executing",
		t->tss.cr2);
	/* fill the struct */
	spin_lock(&lock);
	memcpy(history[sf_index].comm, t->comm, 32);
	history[sf_index].pid = t->pid;
	history[sf_index].address = t->tss.cr2;
	history[sf_index].error_code = t->tss.error_code;
	history[sf_index].uid = t->uid;
	history[sf_index].timestamp = xtime.tv_sec;
	sf_index++;
	sf_index %= HISTORY_SIZE;
	spin_unlock(&lock);
	return SIGHOOK_CONTINUE;
}

static int status_read_proc(char *buf, char **start, off_t offset, int len, int unused)
{
	if (len <= 0)
		return 0;
	return sprintf(buf, "%d\n", sigsegv_hook ? 1 : 0);
}

static int history_read_proc(char *buf, char **start, off_t offset, int len, int unused)
{
	int buf_offset = 0;
	int j, i;
	/* The size of the output MUST be EXTRY_OUTPUT_SIZE */
	const char *format = "%010d %010d %-32s %d %08lx %lu\n";
	char tmp[1024]; /* we fear overruns */

	if (len <= 0)
		return 0;
	spin_lock(&lock);
	i = sf_index;
	for (j = 0; j < HISTORY_SIZE; j++) {
		if (history[i].pid != 0) {
			sprintf(tmp, format,
				history[i].pid, history[i].uid,
				history[i].comm, history[i].error_code,
				history[i].address, history[i].timestamp);
			memcpy(buf+buf_offset, tmp, ENTRY_OUTPUT_SIZE);
			buf_offset += ENTRY_OUTPUT_SIZE;
		}
		i++;
		i %= HISTORY_SIZE;
	}
	spin_unlock(&lock);
	return buf_offset;
}

static int status_write_proc(struct file *file, const char *buf, unsigned long len, void *data)
{
	if (len <= 0)
		return 0;
	if (!access_ok(VERIFY_READ, buf, len))
		return -EFAULT;
	if (buf[0] == '1' && sigsegv_hook == NULL) {
		sigsegv_hook = log_sigsegv;
		printk("SIGSEGV hook registered\n");
	} else if (buf[0] == '0' && sigsegv_hook != NULL) {
		sigsegv_hook = NULL;
		printk("SIGSEGV hook unregistered\n");
	}
	return len;
}

static void clean_history(void)
{
	memset(history, 0, sizeof(struct struct_sigsegv)*HISTORY_SIZE);
}

int init_module(void)
{
	/* we don't register the hook here, so per default logging is off */
	sigsegv_root = create_proc_entry("sys/sigsegv", 0600|S_IFDIR, NULL);
	if (sigsegv_root == NULL)
		goto failed;
	status_entry = create_proc_entry("sys/sigsegv/status", 0600, NULL);
	history_entry = create_proc_entry("sys/sigsegv/history", 0600, NULL);
	if (status_entry == NULL || history_entry == NULL)
		goto failed;

	/* sigsegv_root */
	sigsegv_root->nlink = 1;
	sigsegv_root->uid = 0;
	sigsegv_root->gid = 0;
	sigsegv_root->ops = &proc_dir_inode_operations;

	/* status_entry */
	status_entry->nlink = 1;
	status_entry->uid = 0;
	status_entry->gid = 0;
	status_entry->get_info = &status_read_proc;
	status_entry->write_proc = &status_write_proc;

	/* history_entry */
	history_entry->nlink = 1;
	history_entry->uid = 0;
	history_entry->gid = 0;
	history_entry->get_info = &history_read_proc;

	clean_history();
	printk("SIGSEGV module version " SSVERSION " loaded\n");
	return 0;
failed:
	if (status_entry) remove_proc_entry("sys/sigsegv/status", NULL);
	if (history_entry) remove_proc_entry("sys/sigsegv/history", NULL);
	if (sigsegv_root) remove_proc_entry("sys/sigsegv", NULL);
	printk("SIGSEGV loading failed registering the proc interface\n");
	return -ENOMEM; /* maybe this should be a more generic error */
}

void cleanup_module(void)
{
	/* we must unregister the hook */
	sigsegv_hook = NULL;
	printk("SIGSEGV hook unregistered\n");
	remove_proc_entry("sys/sigsegv/status", NULL);
	remove_proc_entry("sys/sigsegv/history", NULL);
	remove_proc_entry("sys/sigsegv", NULL);
	printk("SIGSEGV module removed\n");
}
