/* segvdump.c
 * dump the segmentation faults history 
 * needs the sigsegv.o module (that relies
 * on sigsegv.patch kernel patch).
 *
 * Copyright (C) 2000 by Salvatore Sanfilippo
 * <antirez@linuxcare.com>
 *
 * This code is under the GPL license
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <pwd.h>
#include <asm/page.h>

#include "ssversion.h"
#include "sscommon.h"

/* globals */
static char *versionstring =	"segvdump " SSVERSION "\n"
				"Copyright (C) 2000 Salvatore Sanfilippo\n"
				"<antirez@linuxcare.com>\n";
static char *procfile = "/proc/sys/sigsegv/history";
static char *openerr = "Opening the proc interface";
static char *readerr = "Can't read from the proc interface";
static char *loadmod = "You must load the sigsegv.o module to use segvdump\n";
static char *permdenied = "Permission denied\n";
static char *formaterror = "format error, wrong segvdump version?\n";

static int opt_raw = 0;		/* raw output */
static int opt_quiet = 0;	/* less verbose */

/* prototypes */
static void print_usage(void);
static void print_version(void);
static void dumpss(void);
static void exit_if_suid(void);
static void dump_human(char *pid, char *uid, char *comm, char *error, char *address, char *time);

int main(int argc, char **argv)
{
	int c;

	while ((c = getopt(argc, argv, "F:hvrq")) != EOF) {
		switch(c) {
		case 'F':
			exit_if_suid();
			procfile = optarg;
			break;
		case 'v':
			print_version();
			break;
		case 'r':
			opt_raw = 1;
			break;
		case 'q':
			opt_quiet = 1;
			break;
		case 'h':
		default:
			print_usage();
			break;
		}
	}
	dumpss();
	return 0;
}

static void dumpss(void)
{
	char buffer[PAGE_SIZE+1];
	int fd;
	int n_read;
	char *pid, *uid, *comm, *error, *address, *time, *p;

	fd = open(procfile, O_RDONLY);
	if (fd == -1) {
		perror(openerr);
		fprintf(stderr, loadmod);
		exit(1);
	}
	n_read = read(fd, buffer, PAGE_SIZE);
	close(fd);
	switch(n_read) {
	case 0:
		printf("history empty\n");
		exit(0);
	case -1:
		perror(readerr);
		exit(1);
	}
	buffer[n_read] = '\0'; /* enforce nul term */
	if (opt_raw) {
		write(fileno(stdout), buffer, n_read);
		exit(0);
	}

	/* The format (not included ^ and $):
	 * ^03521 00000 in.ftpd                          4 000000f7 971303356\n$
	 * every entry is 66 bytes, \n included.
	 */
	if (n_read % ENTRY_OUTPUT_SIZE) {
		fprintf(stderr, formaterror);
		exit(1);
	}
	p = buffer;
	while(n_read) {
		/* write the nul terms */
		p[10] = p[21] = p[54] = p[56] = p[65] = p[75] = '\0';
		pid = p;
		uid = p+11;
		comm = p+22;
		error = p+55;
		address = p+57;
		time = p+66;
		n_read -= ENTRY_OUTPUT_SIZE;
		p += ENTRY_OUTPUT_SIZE;
		dump_human(pid, uid, comm, error, address, time);
	}
	return;
}

/* error_code:
 *      bit 0 == 0 means no page found, 1 means protection fault
 *      bit 1 == 0 means read, 1 means write
 *      bit 2 == 0 means kernel, 1 means user-mode */
static void dump_human(char *pid, char *uid, char *comm, char *error, char *address, char *time)
{
	time_t t = (time_t) strtoul(time, NULL, 10);
	char *date = ctime(&t);
	char *accesstype;
	char *errortype;
	char *errormode;
	char *user;
	char *p;
	struct passwd *pwd;
	char *hformat = "%-20s (pid: %lu, uid: %s) got signal 11\n[%s %s %s] 0x%s at %s\n";
	char *qformat = "%-20s segfaulted [%s %s %s] 0x%s\n";
	int error_code;

	error_code = atoi(error);
	if (error_code & 1)
		errortype = opt_quiet ? "PROTFAULT" : "protection fault";
	else
		errortype = opt_quiet ? "NOPAGE   " : "no page found   ";

	if (error_code & 2)
		accesstype = opt_quiet ? "WRITE    " : "writing             ";
	else
		accesstype = opt_quiet ? "READ/EXEC" : "reading or executing";

	if (error_code & 4)
		errormode = opt_quiet ? "USERMODE  " : "user mode  ";
	else
		errormode = opt_quiet ? "KERNELMODE" : "kernel mode";

	/* print the username instead of the uid if possible */
	pwd = getpwuid((uid_t)strtoul(uid, NULL, 10));
	if (pwd)
		user = pwd->pw_name;
	else
		user = uid;

	/* truncate the trailer spaces in the command */
	p = strchr(comm, ' ');
	if (p) *p = '\0';

	if (opt_quiet)
		printf(qformat, comm, errortype, accesstype, errormode, address);
	else
		printf(hformat, comm, strtoul(pid, NULL, 10), user, errortype,
			accesstype, errormode, address, date);
}

static void print_usage(void)
{
	char *usagestr =
"usage: segvdump [-F file] [-hvrq]\n"
"  -F <file>             use <file> instead of /proc/sigsegv/history\n"
"  -h                    print this help screen and exit\n"
"  -v                    print version information and exit\n"
"  -q                    quiet, less verbose\n";
	printf(usagestr);
	exit(1);
}

static void print_version(void)
{
	printf(versionstring);
	exit(0);
}

static void exit_if_suid(void)
{
	if (getuid() != geteuid()) {
		fprintf(stderr, permdenied);
		exit(1);
	}
}
